# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2018 Laurent Monin
# Copyright (C) 2019 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from collections import (
    OrderedDict,
    namedtuple,
)

from PyQt5 import (
    QtCore,
    QtGui,
    QtNetwork,
    QtWidgets,
)
from PyQt5.QtCore import pyqtSignal

from picard import (
    config,
    log,
)
from picard.util import (
    icontheme,
    natsort,
    restore_method,
    throttle,
)

from picard.ui import PicardDialog
from picard.ui.util import StandardButton


class ResultTable(QtWidgets.QTableWidget):

    def __init__(self, parent):
        super().__init__(parent)
        self.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectRows)
        self.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers)
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Interactive)

        @throttle(1000)  # only emit scrolled signal once per second
        def emit_scrolled(x):
            parent.scrolled.emit()
        self.horizontalScrollBar().valueChanged.connect(emit_scrolled)
        self.verticalScrollBar().valueChanged.connect(emit_scrolled)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

    def prepare(self, headers):
        self.clear()
        self.setColumnCount(len(headers))
        self.setHorizontalHeaderLabels(headers)
        self.setRowCount(0)
        self.setSortingEnabled(False)


class SearchBox(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__(parent)
        self.search_action = QtWidgets.QAction(icontheme.lookup('system-search'),
                                               _("Search"), self)
        self.search_action.setEnabled(False)
        self.search_action.triggered.connect(self.search)
        self.setupUi()

    def focus_in_event(self, event):
        # When focus is on search edit box, need to disable
        # dialog's accept button. This would avoid closing of dialog when user
        # hits enter.
        parent = self.parent()
        if parent.table:
            parent.table.clearSelection()
        parent.accept_button.setEnabled(False)

    def setupUi(self):
        self.layout = QtWidgets.QVBoxLayout(self)
        self.search_row_widget = QtWidgets.QWidget(self)
        self.search_row_layout = QtWidgets.QHBoxLayout(self.search_row_widget)
        self.search_row_layout.setContentsMargins(1, 1, 1, 1)
        self.search_row_layout.setSpacing(1)
        self.search_edit = QtWidgets.QLineEdit(self.search_row_widget)
        self.search_edit.setClearButtonEnabled(True)
        self.search_edit.returnPressed.connect(self.trigger_search_action)
        self.search_edit.textChanged.connect(self.enable_search)
        self.search_edit.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.search_edit.focusInEvent = self.focus_in_event
        self.search_row_layout.addWidget(self.search_edit)
        self.search_button = QtWidgets.QToolButton(self.search_row_widget)
        self.search_button.setAutoRaise(True)
        self.search_button.setDefaultAction(self.search_action)
        self.search_button.setIconSize(QtCore.QSize(22, 22))
        self.search_row_layout.addWidget(self.search_button)
        self.search_row_widget.setLayout(self.search_row_layout)
        self.layout.addWidget(self.search_row_widget)
        self.adv_opt_row_widget = QtWidgets.QWidget(self)
        self.adv_opt_row_layout = QtWidgets.QHBoxLayout(self.adv_opt_row_widget)
        self.adv_opt_row_layout.setAlignment(QtCore.Qt.AlignLeft)
        self.adv_opt_row_layout.setContentsMargins(1, 1, 1, 1)
        self.adv_opt_row_layout.setSpacing(1)
        self.use_adv_search_syntax = QtWidgets.QCheckBox(self.adv_opt_row_widget)
        self.use_adv_search_syntax.setText(_("Use advanced query syntax"))
        self.use_adv_search_syntax.stateChanged.connect(self.update_advanced_syntax_setting)
        self.adv_opt_row_layout.addWidget(self.use_adv_search_syntax)
        self.adv_syntax_help = QtWidgets.QLabel(self.adv_opt_row_widget)
        self.adv_syntax_help.setOpenExternalLinks(True)
        self.adv_syntax_help.setText(_(
            "&#160;(<a href='https://musicbrainz.org/doc/Indexed_Search_Syntax'>"
            "Syntax Help</a>)"))
        self.adv_opt_row_layout.addWidget(self.adv_syntax_help)
        self.adv_opt_row_widget.setLayout(self.adv_opt_row_layout)
        self.layout.addWidget(self.adv_opt_row_widget)
        self.layout.setContentsMargins(1, 1, 1, 1)
        self.layout.setSpacing(1)
        self.setMaximumHeight(60)

    def search(self):
        self.parent().search(self.query)

    def restore_checkbox_state(self):
        self.use_adv_search_syntax.setChecked(config.setting["use_adv_search_syntax"])

    def update_advanced_syntax_setting(self):
        config.setting["use_adv_search_syntax"] = self.use_adv_search_syntax.isChecked()

    def enable_search(self):
        if self.query:
            self.search_action.setEnabled(True)
        else:
            self.search_action.setEnabled(False)

    def trigger_search_action(self):
        if self.search_action.isEnabled():
            self.search_action.trigger()

    def get_query(self):
        return self.search_edit.text()

    def set_query(self, query):
        return self.search_edit.setText(query)

    query = property(get_query, set_query)


Retry = namedtuple("Retry", ["function", "query"])


class SortableTableWidgetItem(QtWidgets.QTableWidgetItem):

    def __init__(self, sort_key):
        super().__init__()
        self.sort_key = sort_key

    def __lt__(self, other):
        return self.sort_key < other.sort_key


class SearchDialog(PicardDialog):

    defaultsize = QtCore.QSize(720, 360)
    autorestore = False
    scrolled = pyqtSignal()

    def __init__(self, parent, accept_button_title, show_search=True, search_type=None):
        super().__init__(parent)
        self.search_results = []
        self.show_search = show_search
        self.search_type = search_type
        self.search_box = None
        self.setupUi(accept_button_title)
        self.restore_state()
        # self.columns has to be an ordered dict, with column name as keys, and
        # matching label as values
        self.columns = None
        self.sorting_enabled = True
        self.create_table()
        self.finished.connect(self.save_state)

    @property
    def columns(self):
        return self.__columns

    @columns.setter
    def columns(self, list_of_tuples):
        if not list_of_tuples:
            list_of_tuples = []
        self.__columns = OrderedDict(list_of_tuples)
        self.__colkeys = list(self.columns.keys())

    @property
    def table_headers(self):
        return list(self.columns.values())

    def colpos(self, colname):
        return self.__colkeys.index(colname)

    def set_table_item(self, row, colname, obj, key, default="", sortkey=None):
        # QVariant remembers the original type of the data
        # matching comparison operator will be used when sorting
        # get() will return a string, force conversion if asked to
        value = obj.get(key, default)
        if sortkey is None:
            sortkey = natsort.natkey(value)
        item = SortableTableWidgetItem(sortkey)
        item.setData(QtCore.Qt.DisplayRole, value)
        pos = self.colpos(colname)
        if pos == 0:
            item.setData(QtCore.Qt.UserRole, row)
        self.table.setItem(row, pos, item)

    def setupUi(self, accept_button_title):
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setObjectName("vertical_layout")
        if self.show_search:
            self.search_box = SearchBox(self)
            self.search_box.setObjectName("search_box")
            self.verticalLayout.addWidget(self.search_box)
        self.center_widget = QtWidgets.QWidget(self)
        self.center_widget.setObjectName("center_widget")
        self.center_layout = QtWidgets.QVBoxLayout(self.center_widget)
        self.center_layout.setObjectName("center_layout")
        self.center_layout.setContentsMargins(1, 1, 1, 1)
        self.center_widget.setLayout(self.center_layout)
        self.verticalLayout.addWidget(self.center_widget)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        if self.show_search and self.search_type:
            self.search_browser_button = QtWidgets.QPushButton(
                _("Search in browser"), self.buttonBox)
            self.buttonBox.addButton(
                self.search_browser_button,
                QtWidgets.QDialogButtonBox.ActionRole)
            self.search_browser_button.clicked.connect(self.search_browser)
        self.accept_button = QtWidgets.QPushButton(
            accept_button_title,
            self.buttonBox)
        self.accept_button.setEnabled(False)
        self.buttonBox.addButton(
            self.accept_button,
            QtWidgets.QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(
            StandardButton(StandardButton.CANCEL),
            QtWidgets.QDialogButtonBox.RejectRole)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.verticalLayout.addWidget(self.buttonBox)

    def add_widget_to_center_layout(self, widget):
        """Update center widget with new child. If child widget exists,
        schedule it for deletion."""
        widget_item = self.center_layout.takeAt(0)
        if widget_item:
            current_widget = widget_item.widget()
            current_widget.hide()
            self.center_layout.removeWidget(current_widget)
            if current_widget != self.table:
                current_widget.deleteLater()
        self.center_layout.addWidget(widget)
        widget.show()

    def show_progress(self):
        progress_widget = QtWidgets.QWidget(self)
        progress_widget.setObjectName("progress_widget")
        layout = QtWidgets.QVBoxLayout(progress_widget)
        text_label = QtWidgets.QLabel(_('<strong>Loading...</strong>'), progress_widget)
        text_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignBottom)
        gif_label = QtWidgets.QLabel(progress_widget)
        movie = QtGui.QMovie(":/images/loader.gif")
        gif_label.setMovie(movie)
        movie.start()
        gif_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop)
        layout.addWidget(text_label)
        layout.addWidget(gif_label)
        layout.setContentsMargins(1, 1, 1, 1)
        progress_widget.setLayout(layout)
        self.add_widget_to_center_layout(progress_widget)

    def show_error(self, error, show_retry_button=False):
        """Display the error string.

        Args:
            error -- Error string
            show_retry_button -- Whether to display retry button or not
        """
        error_widget = QtWidgets.QWidget(self)
        error_widget.setObjectName("error_widget")
        layout = QtWidgets.QVBoxLayout(error_widget)
        error_label = QtWidgets.QLabel(error, error_widget)
        error_label.setWordWrap(True)
        error_label.setAlignment(QtCore.Qt.AlignCenter)
        error_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        layout.addWidget(error_label)
        if show_retry_button:
            retry_widget = QtWidgets.QWidget(error_widget)
            retry_layout = QtWidgets.QHBoxLayout(retry_widget)
            retry_button = QtWidgets.QPushButton(_("Retry"), error_widget)
            retry_button.clicked.connect(self.retry)
            retry_button.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed))
            retry_layout.addWidget(retry_button)
            retry_layout.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop)
            retry_widget.setLayout(retry_layout)
            layout.addWidget(retry_widget)
        error_widget.setLayout(layout)
        self.add_widget_to_center_layout(error_widget)

    def create_table(self):
        self.table = ResultTable(self)
        self.table.verticalHeader().setDefaultSectionSize(100)
        self.table.setSortingEnabled(False)
        self.table.cellDoubleClicked.connect(self.accept)
        self.table.hide()

        def enable_accept_button():
            self.accept_button.setEnabled(True)
        self.table.itemSelectionChanged.connect(
            enable_accept_button)

    def prepare_table(self):
        self.table.prepare(self.table_headers)
        self.restore_table_header_state()

    def show_table(self, sort_column=None, sort_order=QtCore.Qt.DescendingOrder):
        self.add_widget_to_center_layout(self.table)
        self.table.horizontalHeader().setSortIndicatorShown(self.sorting_enabled)
        self.table.setSortingEnabled(self.sorting_enabled)
        if self.sorting_enabled and sort_column:
            self.table.sortItems(self.colpos(sort_column), sort_order)

        self.table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.table.setAlternatingRowColors(True)

    def network_error(self, reply, error):
        error_msg = _("<strong>Following error occurred while fetching results:<br><br></strong>"
                      "Network request error for %s:<br>%s (QT code %d, HTTP code %s)<br>") % (
                          reply.request().url().toString(QtCore.QUrl.RemoveUserInfo),
                          reply.errorString(),
                          error,
                          repr(reply.attribute(
                              QtNetwork.QNetworkRequest.HttpStatusCodeAttribute))
        )
        self.show_error(error_msg, show_retry_button=True)

    def no_results_found(self):
        error_msg = _("<strong>No results found. Please try a different search query.</strong>")
        self.show_error(error_msg)

    def search_browser(self):
        self.tagger.search(self.search_box.query, self.search_type,
                           adv=config.setting["use_adv_search_syntax"], force_browser=True)

    def accept(self):
        if self.table:
            selected_rows = []
            for idx in self.table.selectionModel().selectedRows():
                row = self.table.itemFromIndex(idx).data(QtCore.Qt.UserRole)
                selected_rows.append(row)
            self.accept_event(selected_rows)
        super().accept()

    @restore_method
    def restore_state(self):
        self.restore_geometry()
        if self.show_search:
            self.search_box.restore_checkbox_state()

    @restore_method
    def restore_table_header_state(self):
        header = self.table.horizontalHeader()
        state = config.persist[self.dialog_header_state]
        if state:
            header.restoreState(state)
        header.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
        log.debug("restore_state: %s" % self.dialog_header_state)

    def save_state(self):
        if self.table:
            self.save_table_header_state()

    def save_table_header_state(self):
        state = self.table.horizontalHeader().saveState()
        config.persist[self.dialog_header_state] = state
        log.debug("save_state: %s" % self.dialog_header_state)

    def search_box_text(self, text):
        if self.search_box:
            self.search_box.query = text



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
# Copyright (C) 2008 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from picard import config

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_ratings import Ui_RatingsOptionsPage


class RatingsOptionsPage(OptionsPage):

    NAME = "ratings"
    TITLE = N_("Ratings")
    PARENT = "metadata"
    SORT_ORDER = 20
    ACTIVE = True

    options = [
        config.BoolOption("setting", "enable_ratings", False),
        config.TextOption("setting", "rating_user_email", "users@musicbrainz.org"),
        config.BoolOption("setting", "submit_ratings", True),
        config.IntOption("setting", "rating_steps", 6),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_RatingsOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        self.ui.enable_ratings.setChecked(config.setting["enable_ratings"])
        self.ui.rating_user_email.setText(config.setting["rating_user_email"])
        self.ui.submit_ratings.setChecked(config.setting["submit_ratings"])

    def save(self):
        config.setting["enable_ratings"] = self.ui.enable_ratings.isChecked()
        config.setting["rating_user_email"] = self.ui.rating_user_email.text()
        config.setting["submit_ratings"] = self.ui.submit_ratings.isChecked()


register_options_page(RatingsOptionsPage)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Lukáš Lalinský
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# List of alias locales
ALIAS_LOCALES = {
    'aa': N_('Afar'),
    'aa_DJ': N_('Afar (Djibouti)'),
    'aa_ER': N_('Afar (Eritrea)'),
    'aa_ER_SAAHO': N_('Afar (Eritrea) (Saho)'),
    'aa_ET': N_('Afar (Ethiopia)'),
    'af': N_('Afrikaans'),
    'af_NA': N_('Afrikaans (Namibia)'),
    'af_ZA': N_('Afrikaans (South Africa)'),
    'ak': N_('Akan'),
    'ak_GH': N_('Akan (Ghana)'),
    'sq': N_('Albanian'),
    'sq_AL': N_('Albanian (Albania)'),
    'am': N_('Amharic'),
    'am_ET': N_('Amharic (Ethiopia)'),
    'ar': N_('Arabic'),
    'ar_DZ': N_('Arabic (Algeria)'),
    'ar_BH': N_('Arabic (Bahrain)'),
    'ar_EG': N_('Arabic (Egypt)'),
    'ar_IQ': N_('Arabic (Iraq)'),
    'ar_JO': N_('Arabic (Jordan)'),
    'ar_KW': N_('Arabic (Kuwait)'),
    'ar_LB': N_('Arabic (Lebanon)'),
    'ar_LY': N_('Arabic (Libya)'),
    'ar_MA': N_('Arabic (Morocco)'),
    'ar_OM': N_('Arabic (Oman)'),
    'ar_QA': N_('Arabic (Qatar)'),
    'ar_SA': N_('Arabic (Saudi Arabia)'),
    'ar_SD': N_('Arabic (Sudan)'),
    'ar_SY': N_('Arabic (Syria)'),
    'ar_TN': N_('Arabic (Tunisia)'),
    'ar_AE': N_('Arabic (United Arab Emirates)'),
    'ar_YE': N_('Arabic (Yemen)'),
    'hy': N_('Armenian'),
    'hy_AM': N_('Armenian (Armenia)'),
    'hy_AM_REVISED': N_('Armenian (Armenia) (Revised Orthography)'),
    'as': N_('Assamese'),
    'as_IN': N_('Assamese (India)'),
    'cch': N_('Atsam'),
    'cch_NG': N_('Atsam (Nigeria)'),
    'az': N_('Azerbaijani'),
    'az_AZ': N_('Azerbaijani (Azerbaijan)'),
    'az_Cyrl': N_('Azerbaijani (Cyrillic)'),
    'az_Cyrl_AZ': N_('Azerbaijani (Cyrillic) (Azerbaijan)'),
    'az_Latn': N_('Azerbaijani (Latin)'),
    'az_Latn_AZ': N_('Azerbaijani (Latin) (Azerbaijan)'),
    'eu': N_('Basque'),
    'eu_ES': N_('Basque (Spain)'),
    'be': N_('Belarusian'),
    'be_BY': N_('Belarusian (Belarus)'),
    'bn': N_('Bengali'),
    'bn_BD': N_('Bengali (Bangladesh)'),
    'bn_IN': N_('Bengali (India)'),
    'byn': N_('Blin'),
    'byn_ER': N_('Blin (Eritrea)'),
    'bs': N_('Bosnian'),
    'bs_BA': N_('Bosnian (Bosnia and Herzegovina)'),
    'bg': N_('Bulgarian'),
    'bg_BG': N_('Bulgarian (Bulgaria)'),
    'my': N_('Burmese'),
    'my_MM': N_('Burmese (Myanmar [Burma])'),
    'ca': N_('Catalan'),
    'ca_ES': N_('Catalan (Spain)'),
    'zh': N_('Chinese'),
    'zh_CN': N_('Chinese (China)'),
    'zh_HK': N_('Chinese (Hong Kong)'),
    'zh_MO': N_('Chinese (Macau)'),
    'zh_Hans': N_('Chinese (Simplified Han)'),
    'zh_Hans_CN': N_('Chinese (Simplified Han) (China)'),
    'zh_Hans_HK': N_('Chinese (Simplified Han) (Hong Kong)'),
    'zh_Hans_MO': N_('Chinese (Simplified Han) (Macau)'),
    'zh_Hans_SG': N_('Chinese (Simplified Han) (Singapore)'),
    'zh_SG': N_('Chinese (Singapore)'),
    'zh_TW': N_('Chinese (Taiwan)'),
    'zh_Hant': N_('Chinese (Traditional Han)'),
    'zh_Hant_HK': N_('Chinese (Traditional Han) (Hong Kong)'),
    'zh_Hant_MO': N_('Chinese (Traditional Han) (Macau)'),
    'zh_Hant_TW': N_('Chinese (Traditional Han) (Taiwan)'),
    'cop': N_('Coptic'),
    'kw': N_('Cornish'),
    'kw_GB': N_('Cornish (United Kingdom)'),
    'hr': N_('Croatian'),
    'hr_HR': N_('Croatian (Croatia)'),
    'cs': N_('Czech'),
    'cs_CZ': N_('Czech (Czech Republic)'),
    'da': N_('Danish'),
    'da_DK': N_('Danish (Denmark)'),
    'dv': N_('Divehi'),
    'dv_MV': N_('Divehi (Maldives)'),
    'nl': N_('Dutch'),
    'nl_BE': N_('Dutch (Belgium)'),
    'nl_NL': N_('Dutch (Netherlands)'),
    'dz': N_('Dzongkha'),
    'dz_BT': N_('Dzongkha (Bhutan)'),
    'en': N_('English'),
    'en_AS': N_('English (American Samoa)'),
    'en_AU': N_('English (Australia)'),
    'en_BE': N_('English (Belgium)'),
    'en_BZ': N_('English (Belize)'),
    'en_BW': N_('English (Botswana)'),
    'en_CA': N_('English (Canada)'),
    'en_Dsrt': N_('English (Deseret)'),
    'en_Dsrt_US': N_('English (Deseret) (United States)'),
    'en_GU': N_('English (Guam)'),
    'en_HK': N_('English (Hong Kong)'),
    'en_IN': N_('English (India)'),
    'en_IE': N_('English (Ireland)'),
    'en_JM': N_('English (Jamaica)'),
    'en_MT': N_('English (Malta)'),
    'en_MH': N_('English (Marshall Islands)'),
    'en_NA': N_('English (Namibia)'),
    'en_NZ': N_('English (New Zealand)'),
    'en_MP': N_('English (Northern Mariana Islands)'),
    'en_PK': N_('English (Pakistan)'),
    'en_PH': N_('English (Philippines)'),
    'en_Shaw': N_('English (Shavian)'),
    'en_SG': N_('English (Singapore)'),
    'en_ZA': N_('English (South Africa)'),
    'en_TT': N_('English (Trinidad and Tobago)'),
    'en_UM': N_('English (U.S. Minor Outlying Islands)'),
    'en_VI': N_('English (U.S. Virgin Islands)'),
    'en_GB': N_('English (United Kingdom)'),
    'en_US': N_('English (United States)'),
    'en_US_POSIX': N_('English (United States) (Computer)'),
    'en_ZW': N_('English (Zimbabwe)'),
    'eo': N_('Esperanto'),
    'et': N_('Estonian'),
    'et_EE': N_('Estonian (Estonia)'),
    'ee': N_('Ewe'),
    'ee_GH': N_('Ewe (Ghana)'),
    'ee_TG': N_('Ewe (Togo)'),
    'fo': N_('Faroese'),
    'fo_FO': N_('Faroese (Faroe Islands)'),
    'fil': N_('Filipino'),
    'fil_PH': N_('Filipino (Philippines)'),
    'fi': N_('Finnish'),
    'fi_FI': N_('Finnish (Finland)'),
    'fr': N_('French'),
    'fr_BE': N_('French (Belgium)'),
    'fr_CA': N_('French (Canada)'),
    'fr_FR': N_('French (France)'),
    'fr_LU': N_('French (Luxembourg)'),
    'fr_MC': N_('French (Monaco)'),
    'fr_SN': N_('French (Senegal)'),
    'fr_CH': N_('French (Switzerland)'),
    'fur': N_('Friulian'),
    'fur_IT': N_('Friulian (Italy)'),
    'gaa': N_('Ga'),
    'gaa_GH': N_('Ga (Ghana)'),
    'gl': N_('Galician'),
    'gl_ES': N_('Galician (Spain)'),
    'gez': N_('Geez'),
    'gez_ER': N_('Geez (Eritrea)'),
    'gez_ET': N_('Geez (Ethiopia)'),
    'ka': N_('Georgian'),
    'ka_GE': N_('Georgian (Georgia)'),
    'de': N_('German'),
    'de_AT': N_('German (Austria)'),
    'de_BE': N_('German (Belgium)'),
    'de_DE': N_('German (Germany)'),
    'de_LI': N_('German (Liechtenstein)'),
    'de_LU': N_('German (Luxembourg)'),
    'de_CH': N_('German (Switzerland)'),
    'el_POLYTON': N_('Greek'),
    'el': N_('Greek'),
    'el_CY': N_('Greek (Cyprus)'),
    'el_GR': N_('Greek (Greece)'),
    'gu': N_('Gujarati'),
    'gu_IN': N_('Gujarati (India)'),
    'ha': N_('Hausa'),
    'ha_Arab': N_('Hausa (Arabic)'),
    'ha_Arab_NG': N_('Hausa (Arabic) (Nigeria)'),
    'ha_Arab_SD': N_('Hausa (Arabic) (Sudan)'),
    'ha_GH': N_('Hausa (Ghana)'),
    'ha_Latn': N_('Hausa (Latin)'),
    'ha_Latn_GH': N_('Hausa (Latin) (Ghana)'),
    'ha_Latn_NE': N_('Hausa (Latin) (Niger)'),
    'ha_Latn_NG': N_('Hausa (Latin) (Nigeria)'),
    'ha_NE': N_('Hausa (Niger)'),
    'ha_NG': N_('Hausa (Nigeria)'),
    'ha_SD': N_('Hausa (Sudan)'),
    'haw': N_('Hawaiian'),
    'haw_US': N_('Hawaiian (United States)'),
    'he': N_('Hebrew'),
    'he_IL': N_('Hebrew (Israel)'),
    'hi': N_('Hindi'),
    'hi_IN': N_('Hindi (India)'),
    'hu': N_('Hungarian'),
    'hu_HU': N_('Hungarian (Hungary)'),
    'is': N_('Icelandic'),
    'is_IS': N_('Icelandic (Iceland)'),
    'ig': N_('Igbo'),
    'ig_NG': N_('Igbo (Nigeria)'),
    'id': N_('Indonesian'),
    'id_ID': N_('Indonesian (Indonesia)'),
    'ia': N_('Interlingua'),
    'iu': N_('Inuktitut'),
    'ga': N_('Irish'),
    'ga_IE': N_('Irish (Ireland)'),
    'it': N_('Italian'),
    'it_IT': N_('Italian (Italy)'),
    'it_CH': N_('Italian (Switzerland)'),
    'ja': N_('Japanese'),
    'ja_JP': N_('Japanese (Japan)'),
    'kaj': N_('Jju'),
    'kaj_NG': N_('Jju (Nigeria)'),
    'kl': N_('Kalaallisut'),
    'kl_GL': N_('Kalaallisut (Greenland)'),
    'kam': N_('Kamba'),
    'kam_KE': N_('Kamba (Kenya)'),
    'kn': N_('Kannada'),
    'kn_IN': N_('Kannada (India)'),
    'kk': N_('Kazakh'),
    'kk_Cyrl': N_('Kazakh (Cyrillic)'),
    'kk_Cyrl_KZ': N_('Kazakh (Cyrillic) (Kazakhstan)'),
    'kk_KZ': N_('Kazakh (Kazakhstan)'),
    'km': N_('Khmer'),
    'km_KH': N_('Khmer (Cambodia)'),
    'rw': N_('Kinyarwanda'),
    'rw_RW': N_('Kinyarwanda (Rwanda)'),
    'ky': N_('Kirghiz'),
    'ky_KG': N_('Kirghiz (Kyrgyzstan)'),
    'kok': N_('Konkani'),
    'kok_IN': N_('Konkani (India)'),
    'ko': N_('Korean'),
    'ko_KR': N_('Korean (South Korea)'),
    'kfo': N_('Koro'),
    'kfo_CI': N_('Koro (Côte d’Ivoire)'),
    'kpe': N_('Kpelle'),
    'kpe_GN': N_('Kpelle (Guinea)'),
    'kpe_LR': N_('Kpelle (Liberia)'),
    'ku': N_('Kurdish'),
    'ku_Arab': N_('Kurdish (Arabic)'),
    'ku_Arab_IR': N_('Kurdish (Arabic) (Iran)'),
    'ku_Arab_IQ': N_('Kurdish (Arabic) (Iraq)'),
    'ku_Arab_SY': N_('Kurdish (Arabic) (Syria)'),
    'ku_IR': N_('Kurdish (Iran)'),
    'ku_IQ': N_('Kurdish (Iraq)'),
    'ku_Latn': N_('Kurdish (Latin)'),
    'ku_Latn_TR': N_('Kurdish (Latin) (Turkey)'),
    'ku_SY': N_('Kurdish (Syria)'),
    'ku_TR': N_('Kurdish (Turkey)'),
    'lo': N_('Lao'),
    'lo_LA': N_('Lao (Laos)'),
    'lv': N_('Latvian'),
    'lv_LV': N_('Latvian (Latvia)'),
    'ln': N_('Lingala'),
    'ln_CG': N_('Lingala (Congo - Brazzaville)'),
    'ln_CD': N_('Lingala (Congo - Kinshasa)'),
    'lt': N_('Lithuanian'),
    'lt_LT': N_('Lithuanian (Lithuania)'),
    'nds': N_('Low German'),
    'nds_DE': N_('Low German (Germany)'),
    'mk': N_('Macedonian'),
    'mk_MK': N_('Macedonian (Macedonia)'),
    'ms': N_('Malay'),
    'ms_BN': N_('Malay (Brunei)'),
    'ms_MY': N_('Malay (Malaysia)'),
    'ml': N_('Malayalam'),
    'ml_IN': N_('Malayalam (India)'),
    'mt': N_('Maltese'),
    'mt_MT': N_('Maltese (Malta)'),
    'gv': N_('Manx'),
    'gv_GB': N_('Manx (United Kingdom)'),
    'mr': N_('Marathi'),
    'mr_IN': N_('Marathi (India)'),
    'mo': N_('Moldavian'),
    'mn': N_('Mongolian'),
    'mn_CN': N_('Mongolian (China)'),
    'mn_Cyrl': N_('Mongolian (Cyrillic)'),
    'mn_Cyrl_MN': N_('Mongolian (Cyrillic) (Mongolia)'),
    'mn_MN': N_('Mongolian (Mongolia)'),
    'mn_Mong': N_('Mongolian (Mongolian)'),
    'mn_Mong_CN': N_('Mongolian (Mongolian) (China)'),
    'ne': N_('Nepali'),
    'ne_IN': N_('Nepali (India)'),
    'ne_NP': N_('Nepali (Nepal)'),
    'se': N_('Northern Sami'),
    'se_FI': N_('Northern Sami (Finland)'),
    'se_NO': N_('Northern Sami (Norway)'),
    'nso': N_('Northern Sotho'),
    'nso_ZA': N_('Northern Sotho (South Africa)'),
    'no': N_('Norwegian'),
    'nb': N_('Norwegian Bokmål'),
    'nb_NO': N_('Norwegian Bokmål (Norway)'),
    'nn': N_('Norwegian Nynorsk'),
    'nn_NO': N_('Norwegian Nynorsk (Norway)'),
    'ny': N_('Nyanja'),
    'ny_MW': N_('Nyanja (Malawi)'),
    'oc': N_('Occitan'),
    'oc_FR': N_('Occitan (France)'),
    'or': N_('Oriya'),
    'or_IN': N_('Oriya (India)'),
    'om': N_('Oromo'),
    'om_ET': N_('Oromo (Ethiopia)'),
    'om_KE': N_('Oromo (Kenya)'),
    'ps': N_('Pashto'),
    'ps_AF': N_('Pashto (Afghanistan)'),
    'fa': N_('Persian'),
    'fa_AF': N_('Persian (Afghanistan)'),
    'fa_IR': N_('Persian (Iran)'),
    'pl': N_('Polish'),
    'pl_PL': N_('Polish (Poland)'),
    'pt': N_('Portuguese'),
    'pt_BR': N_('Portuguese (Brazil)'),
    'pt_PT': N_('Portuguese (Portugal)'),
    'pa': N_('Punjabi'),
    'pa_Arab': N_('Punjabi (Arabic)'),
    'pa_Arab_PK': N_('Punjabi (Arabic) (Pakistan)'),
    'pa_Guru': N_('Punjabi (Gurmukhi)'),
    'pa_Guru_IN': N_('Punjabi (Gurmukhi) (India)'),
    'pa_IN': N_('Punjabi (India)'),
    'pa_PK': N_('Punjabi (Pakistan)'),
    'ro': N_('Romanian'),
    'ro_MD': N_('Romanian (Moldova)'),
    'ro_RO': N_('Romanian (Romania)'),
    'ru': N_('Russian'),
    'ru_RU': N_('Russian (Russia)'),
    'ru_UA': N_('Russian (Ukraine)'),
    'sa': N_('Sanskrit'),
    'sa_IN': N_('Sanskrit (India)'),
    'sr_YU': N_('Serbian'),
    'sr': N_('Serbian'),
    'sr_BA': N_('Serbian (Bosnia and Herzegovina)'),
    'sr_Cyrl': N_('Serbian (Cyrillic)'),
    'sr_Cyrl_YU': N_('Serbian (Cyrillic)'),
    'sr_Cyrl_BA': N_('Serbian (Cyrillic) (Bosnia and Herzegovina)'),
    'sr_Cyrl_ME': N_('Serbian (Cyrillic) (Montenegro)'),
    'sr_Cyrl_RS': N_('Serbian (Cyrillic) (Serbia)'),
    'sr_Cyrl_CS': N_('Serbian (Cyrillic) (Serbia and Montenegro)'),
    'sr_Latn': N_('Serbian (Latin)'),
    'sr_Latn_YU': N_('Serbian (Latin)'),
    'sr_Latn_BA': N_('Serbian (Latin) (Bosnia and Herzegovina)'),
    'sr_Latn_ME': N_('Serbian (Latin) (Montenegro)'),
    'sr_Latn_RS': N_('Serbian (Latin) (Serbia)'),
    'sr_Latn_CS': N_('Serbian (Latin) (Serbia and Montenegro)'),
    'sr_ME': N_('Serbian (Montenegro)'),
    'sr_RS': N_('Serbian (Serbia)'),
    'sr_CS': N_('Serbian (Serbia and Montenegro)'),
    'sh': N_('Serbo-Croatian'),
    'sh_YU': N_('Serbo-Croatian'),
    'sh_BA': N_('Serbo-Croatian (Bosnia and Herzegovina)'),
    'sh_CS': N_('Serbo-Croatian (Serbia and Montenegro)'),
    'ii': N_('Sichuan Yi'),
    'ii_CN': N_('Sichuan Yi (China)'),
    'sid': N_('Sidamo'),
    'sid_ET': N_('Sidamo (Ethiopia)'),
    'si': N_('Sinhala'),
    'si_LK': N_('Sinhala (Sri Lanka)'),
    'sk': N_('Slovak'),
    'sk_SK': N_('Slovak (Slovakia)'),
    'sl': N_('Slovenian'),
    'sl_SI': N_('Slovenian (Slovenia)'),
    'so': N_('Somali'),
    'so_DJ': N_('Somali (Djibouti)'),
    'so_ET': N_('Somali (Ethiopia)'),
    'so_KE': N_('Somali (Kenya)'),
    'so_SO': N_('Somali (Somalia)'),
    'nr': N_('South Ndebele'),
    'nr_ZA': N_('South Ndebele (South Africa)'),
    'st': N_('Southern Sotho'),
    'st_LS': N_('Southern Sotho (Lesotho)'),
    'st_ZA': N_('Southern Sotho (South Africa)'),
    'es': N_('Spanish'),
    'es_AR': N_('Spanish (Argentina)'),
    'es_BO': N_('Spanish (Bolivia)'),
    'es_CL': N_('Spanish (Chile)'),
    'es_CO': N_('Spanish (Colombia)'),
    'es_CR': N_('Spanish (Costa Rica)'),
    'es_DO': N_('Spanish (Dominican Republic)'),
    'es_EC': N_('Spanish (Ecuador)'),
    'es_SV': N_('Spanish (El Salvador)'),
    'es_GT': N_('Spanish (Guatemala)'),
    'es_HN': N_('Spanish (Honduras)'),
    'es_MX': N_('Spanish (Mexico)'),
    'es_NI': N_('Spanish (Nicaragua)'),
    'es_PA': N_('Spanish (Panama)'),
    'es_PY': N_('Spanish (Paraguay)'),
    'es_PE': N_('Spanish (Peru)'),
    'es_PR': N_('Spanish (Puerto Rico)'),
    'es_ES': N_('Spanish (Spain)'),
    'es_US': N_('Spanish (United States)'),
    'es_UY': N_('Spanish (Uruguay)'),
    'es_VE': N_('Spanish (Venezuela)'),
    'sw': N_('Swahili'),
    'sw_KE': N_('Swahili (Kenya)'),
    'sw_TZ': N_('Swahili (Tanzania)'),
    'ss': N_('Swati'),
    'ss_ZA': N_('Swati (South Africa)'),
    'ss_SZ': N_('Swati (Swaziland)'),
    'sv': N_('Swedish'),
    'sv_FI': N_('Swedish (Finland)'),
    'sv_SE': N_('Swedish (Sweden)'),
    'gsw': N_('Swiss German'),
    'gsw_CH': N_('Swiss German (Switzerland)'),
    'syr': N_('Syriac'),
    'syr_SY': N_('Syriac (Syria)'),
    'tl': N_('Tagalog'),
    'tg': N_('Tajik'),
    'tg_Cyrl': N_('Tajik (Cyrillic)'),
    'tg_Cyrl_TJ': N_('Tajik (Cyrillic) (Tajikistan)'),
    'tg_TJ': N_('Tajik (Tajikistan)'),
    'ta': N_('Tamil'),
    'ta_IN': N_('Tamil (India)'),
    'trv': N_('Taroko'),
    'trv_TW': N_('Taroko (Taiwan)'),
    'tt': N_('Tatar'),
    'tt_RU': N_('Tatar (Russia)'),
    'te': N_('Telugu'),
    'te_IN': N_('Telugu (India)'),
    'th': N_('Thai'),
    'th_TH': N_('Thai (Thailand)'),
    'bo': N_('Tibetan'),
    'bo_CN': N_('Tibetan (China)'),
    'bo_IN': N_('Tibetan (India)'),
    'tig': N_('Tigre'),
    'tig_ER': N_('Tigre (Eritrea)'),
    'ti': N_('Tigrinya'),
    'ti_ER': N_('Tigrinya (Eritrea)'),
    'ti_ET': N_('Tigrinya (Ethiopia)'),
    'to': N_('Tonga'),
    'to_TO': N_('Tonga (Tonga)'),
    'ts': N_('Tsonga'),
    'ts_ZA': N_('Tsonga (South Africa)'),
    'tn': N_('Tswana'),
    'tn_ZA': N_('Tswana (South Africa)'),
    'tr': N_('Turkish'),
    'tr_TR': N_('Turkish (Turkey)'),
    'kcg': N_('Tyap'),
    'kcg_NG': N_('Tyap (Nigeria)'),
    'ug': N_('Uighur'),
    'ug_Arab': N_('Uighur (Arabic)'),
    'ug_Arab_CN': N_('Uighur (Arabic) (China)'),
    'ug_CN': N_('Uighur (China)'),
    'uk': N_('Ukrainian'),
    'uk_UA': N_('Ukrainian (Ukraine)'),
    'ur': N_('Urdu'),
    'ur_IN': N_('Urdu (India)'),
    'ur_PK': N_('Urdu (Pakistan)'),
    'uz': N_('Uzbek'),
    'uz_AF': N_('Uzbek (Afghanistan)'),
    'uz_Arab': N_('Uzbek (Arabic)'),
    'uz_Arab_AF': N_('Uzbek (Arabic) (Afghanistan)'),
    'uz_Cyrl': N_('Uzbek (Cyrillic)'),
    'uz_Cyrl_UZ': N_('Uzbek (Cyrillic) (Uzbekistan)'),
    'uz_Latn': N_('Uzbek (Latin)'),
    'uz_Latn_UZ': N_('Uzbek (Latin) (Uzbekistan)'),
    'uz_UZ': N_('Uzbek (Uzbekistan)'),
    've': N_('Venda'),
    've_ZA': N_('Venda (South Africa)'),
    'vi': N_('Vietnamese'),
    'vi_VN': N_('Vietnamese (Vietnam)'),
    'wal': N_('Walamo'),
    'wal_ET': N_('Walamo (Ethiopia)'),
    'cy': N_('Welsh'),
    'cy_GB': N_('Welsh (United Kingdom)'),
    'wo': N_('Wolof'),
    'wo_Latn': N_('Wolof (Latin)'),
    'wo_Latn_SN': N_('Wolof (Latin) (Senegal)'),
    'wo_SN': N_('Wolof (Senegal)'),
    'xh': N_('Xhosa'),
    'xh_ZA': N_('Xhosa (South Africa)'),
    'yo': N_('Yoruba'),
    'yo_NG': N_('Yoruba (Nigeria)'),
    'zu': N_('Zulu'),
    'zu_ZA': N_('Zulu (South Africa)'),
}



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AboutOptionsPage(object):
    def setupUi(self, AboutOptionsPage):
        AboutOptionsPage.setObjectName("AboutOptionsPage")
        AboutOptionsPage.resize(171, 137)
        self.vboxlayout = QtWidgets.QVBoxLayout(AboutOptionsPage)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.scrollArea = QtWidgets.QScrollArea(AboutOptionsPage)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setFrameShadow(QtWidgets.QFrame.Plain)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 171, 137))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setContentsMargins(9, 9, 9, 9)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label.setText("")
        self.label.setAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.vboxlayout.addWidget(self.scrollArea)

        self.retranslateUi(AboutOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(AboutOptionsPage)

    def retranslateUi(self, AboutOptionsPage):
        pass




# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
# Copyright (C) 2006-2007 Lukáš Lalinský
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
from collections import namedtuple
from collections.abc import (
    Iterable,
    MutableMapping,
)

from PyQt5.QtCore import QObject

from picard import config
from picard.mbjson import (
    artist_credit_from_node,
    get_score,
)
from picard.plugin import (
    PluginFunctions,
    PluginPriority,
)
from picard.similarity import similarity2
from picard.util import linear_combination_of_weights
from picard.util.imagelist import ImageList
from picard.util.tags import PRESERVED_TAGS


MULTI_VALUED_JOINER = '; '

# lengths difference over this number of milliseconds will give a score of 0.0
# equal lengths will give a score of 1.0
# example
# a     b     score
# 20000 0     0.333333333333
# 20000 10000 0.666666666667
# 20000 20000 1.0
# 20000 30000 0.666666666667
# 20000 40000 0.333333333333
# 20000 50000 0.0
LENGTH_SCORE_THRES_MS = 30000

SimMatchTrack = namedtuple('SimMatchTrack', 'similarity releasegroup release track')
SimMatchRelease = namedtuple('SimMatchRelease', 'similarity release')


def weights_from_release_type_scores(parts, release, release_type_scores,
                                     weight_release_type=1):
    # This function generates a score that determines how likely this release will be selected in a lookup.
    # The score goes from 0 to 1 with 1 being the most likely to be chosen and 0 the least likely
    # This score is based on the preferences of release-types found in this release
    # This algorithm works by taking the scores of the primary type (and secondary if found) and averages them
    # If no types are found, it is set to the score of the 'Other' type or 0.5 if 'Other' doesnt exist
    # It appends (score, weight_release_type) to passed parts list

    type_scores = dict(release_type_scores)
    score = 0.0
    if 'release-group' in release and 'primary-type' in release['release-group']:
        types_found = [release['release-group']['primary-type']]
        if 'secondary-types' in release['release-group']:
            types_found += release['release-group']['secondary-types']
        other_score = type_scores.get('Other', 0.5)
        for release_type in types_found:
            score += type_scores.get(release_type, other_score)
        score /= len(types_found)
    parts.append((score, weight_release_type))


def weights_from_preferred_countries(parts, release,
                                     preferred_countries,
                                     weight):
    total_countries = len(preferred_countries)
    if total_countries:
        score = 0.0
        if "country" in release:
            try:
                i = preferred_countries.index(release['country'])
                score = float(total_countries - i) / float(total_countries)
            except ValueError:
                pass
        parts.append((score, weight))


def weights_from_preferred_formats(parts, release, preferred_formats, weight):
    total_formats = len(preferred_formats)
    if total_formats and 'media' in release:
        score = 0.0
        subtotal = 0
        for medium in release['media']:
            if "format" in medium:
                try:
                    i = preferred_formats.index(medium['format'])
                    score += float(total_formats - i) / float(total_formats)
                except ValueError:
                    pass
                subtotal += 1
        if subtotal > 0:
            score /= subtotal
        parts.append((score, weight))


class Metadata(MutableMapping):

    """List of metadata items with dict-like access."""

    __weights = [
        ('title', 22),
        ('artist', 6),
        ('album', 12),
        ('tracknumber', 6),
        ('totaltracks', 5),
    ]

    multi_valued_joiner = MULTI_VALUED_JOINER

    def __init__(self, *args, deleted_tags=None, images=None, length=None, **kwargs):
        self._store = dict()
        self.deleted_tags = set()
        self.length = 0
        self.images = ImageList()
        self.has_common_images = True

        d = dict(*args, **kwargs)
        for k, v in d.items():
            self[k] = v
        if images is not None:
            for image in images:
                self.images.append(image)
        if deleted_tags is not None:
            for tag in deleted_tags:
                del self[tag]
        if length is not None:
            self.length = int(length)

    def __bool__(self):
        return bool(len(self))

    def __len__(self):
        return len(self._store) + len(self.images)

    @staticmethod
    def length_score(a, b):
        return (1.0 - min(abs(a - b),
                LENGTH_SCORE_THRES_MS) / float(LENGTH_SCORE_THRES_MS))

    def compare(self, other, ignored=None):
        parts = []
        if ignored is None:
            ignored = []

        if self.length and other.length and '~length' not in ignored:
            score = self.length_score(self.length, other.length)
            parts.append((score, 8))

        for name, weight in self.__weights:
            if name in ignored:
                continue
            a = self[name]
            b = other[name]
            if a and b:
                if name in ('tracknumber', 'totaltracks'):
                    try:
                        ia = int(a)
                        ib = int(b)
                    except ValueError:
                        ia = a
                        ib = b
                    score = 1.0 - (int(ia != ib))
                else:
                    score = similarity2(a, b)
                parts.append((score, weight))
            elif (a and name in other.deleted_tags
                  or b and name in self.deleted_tags):
                parts.append((0, weight))
        return linear_combination_of_weights(parts)

    def compare_to_release(self, release, weights):
        """
        Compare metadata to a MusicBrainz release. Produces a probability as a
        linear combination of weights that the metadata matches a certain album.
        """
        parts = self.compare_to_release_parts(release, weights)
        sim = linear_combination_of_weights(parts) * get_score(release)
        return SimMatchRelease(similarity=sim, release=release)

    def compare_to_release_parts(self, release, weights):
        parts = []
        if "album" in self:
            b = release['title']
            parts.append((similarity2(self["album"], b), weights["album"]))

        if "albumartist" in self and "albumartist" in weights:
            a = self["albumartist"]
            b = artist_credit_from_node(release['artist-credit'])[0]
            parts.append((similarity2(a, b), weights["albumartist"]))

        try:
            a = int(self["totaltracks"])
            b = release['track-count']
            score = 0.0 if a > b else 0.3 if a < b else 1.0
            parts.append((score, weights["totaltracks"]))
        except (ValueError, KeyError):
            pass

        weights_from_preferred_countries(parts, release,
                                         config.setting["preferred_release_countries"],
                                         weights["releasecountry"])

        weights_from_preferred_formats(parts, release,
                                       config.setting["preferred_release_formats"],
                                       weights["format"])

        if "releasetype" in weights:
            weights_from_release_type_scores(parts, release,
                                             config.setting["release_type_scores"],
                                             weights["releasetype"])

        rg = QObject.tagger.get_release_group_by_id(release['release-group']['id'])
        if release['id'] in rg.loaded_albums:
            parts.append((1.0, 6))

        return parts

    def compare_to_track(self, track, weights):
        parts = []

        if 'title' in self:
            a = self['title']
            b = track.get('title', '')
            parts.append((similarity2(a, b), weights["title"]))

        if 'artist' in self:
            a = self['artist']
            artist_credits = track.get('artist-credit', [])
            b = artist_credit_from_node(artist_credits)[0]
            parts.append((similarity2(a, b), weights["artist"]))

        a = self.length
        if a > 0 and 'length' in track:
            b = track['length']
            score = self.length_score(a, b)
            parts.append((score, weights["length"]))

        releases = []
        if "releases" in track:
            releases = track['releases']

        search_score = get_score(track)
        if not releases:
            sim = linear_combination_of_weights(parts) * search_score
            return SimMatchTrack(similarity=sim, releasegroup=None, release=None, track=track)

        if 'isvideo' in weights:
            metadata_is_video = self['~video'] == '1'
            track_is_video = track.get('video', False)
            score = 1 if metadata_is_video == track_is_video else 0
            parts.append((score, weights['isvideo']))

        result = SimMatchTrack(similarity=-1, releasegroup=None, release=None, track=None)
        for release in releases:
            release_parts = self.compare_to_release_parts(release, weights)
            sim = linear_combination_of_weights(parts + release_parts) * search_score
            if sim > result.similarity:
                rg = release['release-group'] if "release-group" in release else None
                result = SimMatchTrack(similarity=sim, releasegroup=rg, release=release, track=track)
        return result

    def copy(self, other, copy_images=True):
        self.clear()
        self._update_from_metadata(other, copy_images)

    def update(self, *args, **kwargs):
        one_arg = len(args) == 1
        if one_arg and isinstance(args[0], self.__class__):
            self._update_from_metadata(args[0])
        elif one_arg and isinstance(args[0], MutableMapping):
            # update from MutableMapping (ie. dict)
            for k, v in args[0].items():
                self[k] = v
        elif args or kwargs:
            # update from a dict-like constructor parameters
            for k, v in dict(*args, **kwargs).items():
                self[k] = v
        else:
            # no argument, raise TypeError to mimic dict.update()
            raise TypeError("descriptor 'update' of '%s' object needs an argument" % self.__class__.__name__)

    def _update_from_metadata(self, other, copy_images=True):
        for k, v in other.rawitems():
            self.set(k, v[:])

        for tag in other.deleted_tags:
            del self[tag]

        if copy_images and other.images:
            self.images = other.images.copy()
        if other.length:
            self.length = other.length

    def clear(self):
        self._store.clear()
        self.images = ImageList()
        self.length = 0
        self.clear_deleted()

    def clear_deleted(self):
        self.deleted_tags = set()

    @staticmethod
    def normalize_tag(name):
        return name.rstrip(':')

    def getall(self, name):
        return self._store.get(self.normalize_tag(name), [])

    def getraw(self, name):
        return self._store[self.normalize_tag(name)]

    def get(self, key, default=None):
        values = self._store.get(self.normalize_tag(key), None)
        if values:
            return self.multi_valued_joiner.join(values)
        else:
            return default

    def __getitem__(self, name):
        return self.get(self.normalize_tag(name), '')

    def set(self, name, values):
        name = self.normalize_tag(name)
        if isinstance(values, str) or not isinstance(values, Iterable):
            values = [values]
        values = [str(value) for value in values if value or value == 0]
        if values:
            self._store[name] = values
            self.deleted_tags.discard(name)
        elif name in self._store:
            del self[name]

    def __setitem__(self, name, values):
        self.set(self.normalize_tag(name), values)

    def __contains__(self, name):
        return self._store.__contains__(self.normalize_tag(name))

    def __delitem__(self, name):
        name = self.normalize_tag(name)
        try:
            del self._store[name]
        except KeyError:
            pass
        finally:
            self.deleted_tags.add(name)

    def add(self, name, value):
        if value or value == 0:
            name = self.normalize_tag(name)
            self._store.setdefault(name, []).append(str(value))
            self.deleted_tags.discard(name)

    def add_unique(self, name, value):
        name = self.normalize_tag(name)
        if value not in self.getall(name):
            self.add(name, value)

    def delete(self, name):
        """Deprecated: use del directly"""
        del self[self.normalize_tag(name)]

    def __iter__(self):
        return iter(self._store)

    def items(self):
        for name, values in self._store.items():
            for value in values:
                yield name, value

    def rawitems(self):
        """Returns the metadata items.

        >>> m.rawitems()
        [("key1", ["value1", "value2"]), ("key2", ["value3"])]
        """
        return self._store.items()

    def apply_func(self, func):
        for name, values in list(self.rawitems()):
            if name not in PRESERVED_TAGS:
                self[name] = [func(value) for value in values]

    def strip_whitespace(self):
        """Strip leading/trailing whitespace.

        >>> m = Metadata()
        >>> m["foo"] = "  bar  "
        >>> m["foo"]
        "  bar  "
        >>> m.strip_whitespace()
        >>> m["foo"]
        "bar"
        """
        self.apply_func(str.strip)

    def __repr__(self):
        return "%s(%r, deleted_tags=%r, length=%r, images=%r)" % (self.__class__.__name__, self._store, self.deleted_tags, self.length, self.images)

    def __str__(self):
        return ("store: %r\ndeleted: %r\nimages: %r\nlength: %r" % (self._store, self.deleted_tags, [str(img) for img in self.images], self.length))


_album_metadata_processors = PluginFunctions(label='album_metadata_processors')
_track_metadata_processors = PluginFunctions(label='track_metadata_processors')


def register_album_metadata_processor(function, priority=PluginPriority.NORMAL):
    """Registers new album-level metadata processor."""
    _album_metadata_processors.register(function.__module__, function, priority)


def register_track_metadata_processor(function, priority=PluginPriority.NORMAL):
    """Registers new track-level metadata processor."""
    _track_metadata_processors.register(function.__module__, function, priority)


def run_album_metadata_processors(album_object, metadata, release):
    _album_metadata_processors.run(album_object, metadata, release)


def run_track_metadata_processors(album_object, metadata, track, release=None):
    _track_metadata_processors.run(album_object, metadata, track, release)
